import java.lang.*;
import java.util.*;
import org.w3c.dom.*;
import java.io.*;

import javax.xml.parsers.*;
import org.xml.sax.*;

public class DataGraph {
    /*
    static private int SHORT_LENGTH = 2;
    static private int INTEGER_LENGTH = 4;
    static private int LONG_LENGTH = 8;
    */
    static private int BUFFER_SIZE = 1024*8;

    private String _data_file_name = null;
    private String _index_directory = null;
    private long _eof_offset = 0;
    private long _root_offset = 0;
    private RandomAccessFile _data_file = null;

    private DataOutputStream _data_out_stream = null;

    private long _num_cross_edges = 0;

    // needed for merging parent info
    private Vector _parent_out_streams = null;
    
    // needed to create dataguide
    private Vector _dg_nodes = null;
    private Hashtable _path_dgindex_map = null;
    private Vector _dg_out_edges = null;
    private Vector _dg_in_edges = null;
    private int _root_dg_node_index = 0;

    // should the index file be saved at the end
    private boolean _write_flag = false;

    Hashtable chain = new Hashtable();
    Stack inStack = new Stack();
    Vector leaves = new Vector();
    Vector internal = new Vector();


    /**
     * Constructor.
     *
     * @param index_directory_ directory where index files are to be created
     * @param data_file_name_  name of file to be indexed
     * @param write_flag_      boolean flag that specifies whether the index files have to be 
     *                         preserved at the termination of the JVM
     */

    DataGraph(String index_directory_, String data_file_name_, boolean write_flag_) {
	_write_flag = write_flag_; 
	_index_directory = index_directory_;

	_data_file_name = index_directory_+data_file_name_+".out";

	File index_file = new File(_data_file_name);
	if(write_flag_ && index_file.exists()) {	    
	    System.out.println("Using existing index files");
	    try {
		_data_file = new RandomAccessFile(_data_file_name, "r");
	    }
	    catch(FileNotFoundException fnfe) {
		System.out.println("FileNotFoundException during data file creation " + fnfe.toString());
		System.exit(0);
	    }
	    catch(IOException ioe) {
		System.out.println("IOException during data file creation " + ioe.toString());
		System.exit(0);
	    }
	    RandomAccessFile data_file = null;

	    try {
		data_file = new RandomAccessFile(_data_file_name+".idx", "r");
	    }
	    catch(FileNotFoundException fnfe) {
		System.out.println("FileNotFoundException during data file creation " + fnfe.toString());
		System.exit(0);
	    }
	    catch(IOException ioe) {
		System.out.println("IOException during data file creation " + ioe.toString());
		System.exit(0);
	    }
	    try {
		_eof_offset = data_file.readLong();
		_root_offset = data_file.readLong();
		data_file.close();
	    }
	    catch(IOException ioe) {
		System.out.println("IOException during data file read " + ioe.toString());
		System.exit(0);
	    }
	    data_file = null;
	    return;
	}

	_parent_out_streams = new Vector();
	try {
	    _data_out_stream = new DataOutputStream(new BufferedOutputStream(new FileOutputStream(_data_file_name), BUFFER_SIZE));
	}
	catch(FileNotFoundException fnfe) {
	    System.out.println("FileNotFoundException during data file creation " + fnfe.toString());
	    System.exit(0);
	}
	catch(IOException ioe) {
	    System.out.println("IOException during data file creation " + ioe.toString());
	    System.exit(0);
	}
	_path_dgindex_map = new Hashtable();
	_dg_nodes = new Vector();
	_dg_out_edges  = new Vector();
	_dg_in_edges = new Vector();

	parseFile(data_file_name_);	
	if(MyFrame._debug_mode > 8) {
	    System.out.println("eof_offset = " + _eof_offset);
	    System.out.println("root_offset = " + _root_offset);
	}

	try {
	    _data_out_stream.close();
	    Enumeration e = _parent_out_streams.elements();
	    while(e.hasMoreElements()) {
		DataOutputStream out_stream = (DataOutputStream)(e.nextElement());
		out_stream.close();
	    }	    
	}
	catch(IOException ioe) {
	    System.out.println("IOException during data file creation " + ioe.toString());
	    System.exit(0);
	}
	_data_out_stream = null;
	if(MyFrame._debug_mode > 0) {
	    System.out.println("Updating parent information");
	}
	long start_time = System.currentTimeMillis();
	updateParents();
	long time_taken = System.currentTimeMillis()-start_time;
	if(MyFrame._debug_mode > 0) {
	    System.out.println("Updated parent information; Time taken = "+time_taken+" milliseconds");
	}

	DataOutputStream data_file = null;	
	String data_file_name = null;
	try {
	    data_file_name = _data_file_name+".idx";
	    data_file = new DataOutputStream(new BufferedOutputStream(new FileOutputStream(data_file_name), BUFFER_SIZE));
	}
	catch(FileNotFoundException fnfe) {
	    System.out.println("FileNotFoundException during data file creation " + fnfe.toString());
	    System.exit(0);
	}
	catch(IOException ioe) {
	    System.out.println("IOException during data file creation " + ioe.toString());
	    System.exit(0);
	}

	try {
	    data_file.writeLong(_eof_offset);
	    data_file.writeLong(_root_offset);
	    data_file.close();
	}
	catch(IOException ioe) {
	    System.out.println("IOException during data file write " + ioe.toString());
	    System.exit(0);
	}	
    }
    
    /**
     * function to perform cleanup at the termination of the JVM.
     */

    protected void finalize() throws Throwable {
	if(!_write_flag) {
	    System.out.print("Removing index file "+_data_file_name);
	    File index_file = new File(_data_file_name);
	    if(index_file.delete() == false) {
		System.out.println("file delete failed; file = " +_data_file_name);
	    }
	    File index2_file = new File(_data_file_name+".idx");
	    if(index2_file.delete() == false) {
		System.out.println("file delete failed; file = " +_data_file_name+".idx");
	    }
	}
    }

    /**
     * function to return the flag that specifies whether index files have to be preserved.
     *
     * @return true if index files are preserved at the termination of the JVM
     *         false if index files are deleted at the termination of the JVM
     */

    public boolean getWriteFlag() {
	return _write_flag;
    }

    /**
     * function that returns the name of the XML file that this data graph encodes.
     *
     * @return the String object representing the name of the file that this data graph encodes.
     */

    public String getFileName() {
	return _data_file_name;
    }

    /**
     * function that returns the name of the directory where index files are written.
     *
     * @return the String object representing the name of the directory where the index files for this 
     *         data graph are written.
     */

    public String getIndexDirectory() {
	return _index_directory;
    }

    /**
     * function to update the perform patching of parent pointers.
     * 
     */
    private void updateParents() {
	int max_depth = _parent_out_streams.size()-1;
	_parent_out_streams = null;
	while(max_depth >= 0) {
	    if(MyFrame._debug_mode > 0) {
		System.out.println("\tProcessing depth "+max_depth);
	    }
	    // open the index_file for input
	    DataInputStream index_in_stream = null;
	    try {
		index_in_stream = new DataInputStream(new BufferedInputStream(new FileInputStream(_data_file_name), BUFFER_SIZE));
	    }
	    catch(FileNotFoundException fnfe) {
		System.out.println("File not found, file = " + _data_file_name + " "+ fnfe.toString());
	    }
	    
	    // open the output stream
	    String temp_file = new String(_data_file_name+".tmp");
	    DataOutputStream index_out_stream = null;
	    
	    try {
		index_out_stream = new DataOutputStream(new BufferedOutputStream(new FileOutputStream(temp_file), BUFFER_SIZE));
	    }
	    catch(FileNotFoundException fnfe) {
		System.out.println("File not found, file = " + temp_file + " "+ fnfe.toString());
	    }
	    
	    // open the patch file for this depth
	    DataInputStream patch_stream = null;
	    try {
		patch_stream = new DataInputStream(new BufferedInputStream(new FileInputStream(String.valueOf(max_depth)), 8192));
	    }
	    catch(FileNotFoundException fnfe) {
		System.out.println("File not found, file = " + max_depth + " "+ fnfe.toString());
	    }
	    
	    // merge the index_file and patch file; write result to output stream
	    int counter = 0;	    
	    boolean read_parent = true;
	    boolean done_flag = false;
	    long child_offset = -1, parent_offset = -1;
	    while(!done_flag) {
		try {
		    if(read_parent) {
			child_offset = patch_stream.readLong();
			parent_offset = patch_stream.readLong();
			if(MyFrame._debug_mode > 8) {
			    System.out.println("Child = " + child_offset + ", parent = " + parent_offset);
			}
			counter++;
			read_parent = false;
		    }
		}
		catch(IOException ioe) {
		    if(MyFrame._debug_mode > 5) {
			System.out.println("Number of edges = " + counter + "; EOF during write " + ioe.toString());
		    }
		    read_parent = false;
		}

		try {
		    int curr_child = index_out_stream.size();
		    if(curr_child == (int)(child_offset)) {
			int num_parents = index_in_stream.readInt(); // num parents
			index_out_stream.writeInt(num_parents);
			long curr_parent = index_in_stream.readLong(); // first parent
			if(curr_parent != -1) System.out.println("Found child; Current parent = "+curr_parent+", New parent = "+parent_offset);
			index_out_stream.writeLong(parent_offset);
			for(int i=0; i<(num_parents-1); i++) {
			    index_out_stream.writeLong(index_in_stream.readLong());
			}
			index_out_stream.writeInt(index_in_stream.readInt()); // path
			index_out_stream.writeUTF(index_in_stream.readUTF()); // label
			index_out_stream.writeUTF(index_in_stream.readUTF()); // data
			int num_children = index_in_stream.readInt(); // num children
			index_out_stream.writeInt(num_children);
			for(int i=0; i<num_children; i++) {
			    index_out_stream.writeLong(index_in_stream.readLong());
			}
			read_parent = true;			
		    }
		    else {
			int num_parents = index_in_stream.readInt(); // num parents
			index_out_stream.writeInt(num_parents);
			for(int i=0; i<num_parents; i++) {
			    long curr_parent = index_in_stream.readLong();
			    if(max_depth == 0 && curr_parent == -1) 
				System.out.println("fatal Error: Not found child; curr_parent = "+curr_parent+", curr_child = "+curr_child);
			    index_out_stream.writeLong(curr_parent);
			}
			index_out_stream.writeInt(index_in_stream.readInt()); // path
			index_out_stream.writeUTF(index_in_stream.readUTF()); // label
			index_out_stream.writeUTF(index_in_stream.readUTF()); // data
			int num_children = index_in_stream.readInt(); // num children
			index_out_stream.writeInt(num_children);
			for(int i=0; i<num_children; i++) {
			    index_out_stream.writeLong(index_in_stream.readLong());
			}			
		    }
		}
		catch(IOException ioe) {
		    if(MyFrame._debug_mode > 5) {
			System.out.println(max_depth + ": Merge complete " + ioe.toString());
		    }
		    done_flag = true;
		}		
	    }
	    
	    try {
		index_out_stream.close();
		index_in_stream.close();
		patch_stream.close();
	    }
	    catch(IOException ioe) {
		System.out.println("ioexception during stream close" + ioe.toString());
		System.exit(0);
	    }

	    // remove the patch file
	    File patch_file_obj = new File(String.valueOf(max_depth));
	    if(patch_file_obj.delete() == false) {
		System.out.println("file delete failed; file = "+max_depth);
		System.exit(0);
	    }
	    
	    // rename the temp file
	    File temp_file_obj = new File(temp_file);
	    if(temp_file_obj.renameTo(new File(_data_file_name)) == false) {
		System.out.println("rename failed (iteration = "+max_depth+")");
		System.exit(0);
	    }

	    max_depth--;
	}	
    }
    
    /**
     * function to add a cross (non-tree) edge between two data graph nodes.
     *
     * @param id_string_ identifier of the data node that is the source of the cross edge
     * @param idref_string_ identifier of the data node that is the terminus of the cross edge
     */
    public void addCrossEdge(String id_string_, String idref_string_) {
	if(MyFrame._debug_mode > 5) {
	    System.out.println("Cross: "+id_string_ + " " +idref_string_);
	}
	_num_cross_edges++;	
    }

    /**
     * function to add a data node to the index file.
     *
     * @param data_node_ Node to be added to the index file
     * @param depth_ Depth of the node in the XML file (root is at depth 0)
     */
    public void addDataNode(DataNode data_node_, int depth_) {
	// write data node to data file
	long curr_offset = _eof_offset;
	try {
	    Vector parents = data_node_.getParentOffsets();
	    int num_parents = parents.size();
	    _data_out_stream.writeInt(num_parents); 
	    Enumeration parents_enum = parents.elements();
	    while(parents_enum.hasMoreElements()) {
		Long parent_offset = ((Long)(parents_enum.nextElement()));
		_data_out_stream.writeLong(parent_offset.longValue());
	    }

	    if(MyFrame._debug_mode > 8) {
		System.out.println("Path = \""+data_node_.getPath()+"\"");
		System.out.println("Data = \""+data_node_.getData()+"\"");
		System.out.println("Labl = \""+data_node_.getLabel()+"\"");
	    }
	    _data_out_stream.writeInt(data_node_.getPath());
	    _data_out_stream.writeUTF(data_node_.getLabel());
	    _data_out_stream.writeUTF(data_node_.getData());
	    
	    Vector children = data_node_.getChildOffsets();
	    int num_children = children.size();
	    _data_out_stream.writeInt(num_children); 
	    if(MyFrame._debug_mode > 8) {
		System.out.println("Numc = "+num_children);
	    }
	    Enumeration children_enum = children.elements();
	    while(children_enum.hasMoreElements()) {
		Long child_offset = (Long)(children_enum.nextElement());
		_data_out_stream.writeLong(child_offset.longValue());
	    }	   
	}
	catch(IOException ioe) {
	    System.out.println("IOException during write " + ioe.toString());
	    System.exit(0);
	}
	
	
	Vector children = data_node_.getChildOffsets();
	int num_children = children.size();
	
	_eof_offset = _data_out_stream.size();

	// update parent pointers for children	
	Enumeration children_enum = children.elements();
	while(children_enum.hasMoreElements()) {
	    Long child_offset = (Long)(children_enum.nextElement());	   
	    DataOutputStream out_stream = (DataOutputStream)(_parent_out_streams.elementAt(depth_));
	    try {
		out_stream.writeLong(child_offset.longValue());
		out_stream.writeLong(curr_offset);
	    }
	    catch(IOException ioe) {
		System.out.println("IOException during write " + ioe.toString());
		System.exit(0);
	    }
	}
    }

    /**
     * function to create an output stream to write patch (parent) information.
     *
     * @param depth_ depth for which the parent output stream is to be created
     */

    public void createParentOutStream(Integer depth_) {
	DataOutputStream object_out_stream = null;
	try {
	    object_out_stream = new DataOutputStream(new BufferedOutputStream(new FileOutputStream(depth_.toString())));
	}
	catch(IOException ioe) {
	    System.out.println("IOException during data file creation " + ioe.toString());
	    System.exit(0);
	}
	
	_parent_out_streams.addElement(object_out_stream);
    }

    
    /**
     * function that checks if the data node at offset offset1_ is an ancestor of the data node at offset offset2_
     * 
     * @param offset1_ offset of first data node
     * @param offset2_ offset of second data node
     * @return true if first data node is an ancestor of second data node
     */
    public boolean is_ancestor(long offset1_, long offset2_) {
	DataNode curr_data_node = getDataNodeNoChildren(offset2_);
	Vector  parents = curr_data_node.getParentOffsets();
	Enumeration e = parents.elements();
	while(e.hasMoreElements()) {
	    Long parent_offset = (Long)(e.nextElement());
	    if(parent_offset.longValue() == offset1_) return true;
	}

	e = parents.elements();
	while(e.hasMoreElements()) {
	    Long parent_offset = (Long)(e.nextElement());
	    if(is_ancestor(offset1_, parent_offset.longValue())) return true;
	}
	return false;
    }

    /**
     * function that returns the data node object whose offset is the argument offset
     * 
     * @param offset_ offset of a node
     * @return DataNode object whose offset is the argument offset
     */     
    public DataNode getDataNodeNoChildren(long offset_) {
	DataNode return_node = null;
	if(_data_file == null) {
	    try {
		_data_file = new RandomAccessFile(_data_file_name, "r");
	    }
	    catch(FileNotFoundException fnfe) {
		System.out.println("FileNotFoundException during data file creation " + fnfe.toString());
		System.exit(0);
	    }
	    catch(IOException ioe) {
		System.out.println("IOException during data file creation " + ioe.toString());
		System.exit(0);
	    }
	}
	if(offset_ < _eof_offset) {
	    try {
		_data_file.seek(offset_);
	    }
	    catch(IOException ioe) {
		System.out.println("IOException in seek() " + ioe.toString());
		System.exit(0);
	    }
	    try {
		return_node = new DataNode(new String(""));

		int num_parents = _data_file.readInt(); // num_parents
		if(MyFrame._debug_mode > 8) {
		    System.out.println("Num parents = "+num_parents);
		}
		for(int c=0; c<num_parents; c++) {
 		    long parent_offset = _data_file.readLong(); // parent_offset
		    return_node.addParentOffset(parent_offset);
		}

		int path = _data_file.readInt();
		if(MyFrame._debug_mode > 8) {
		    System.out.println("Path = "+path);
		}
		return_node.setPath(path);

		String label = _data_file.readUTF();
		if(MyFrame._debug_mode > 8) {
		    System.out.println("Label = "+label);
		}
		return_node.setLabel(label);

		String data = _data_file.readUTF();
		if(MyFrame._debug_mode > 8) {
		    System.out.println("Data = "+data);
		}
		return_node.setData(data);

		int num_children = _data_file.readInt(); // num_children
		if(MyFrame._debug_mode > 8) {
		    System.out.println("Num children = "+num_children);
		}

		return_node.setChildrenLocation(_data_file_name, _data_file.getFilePointer(), num_children);
	    }
	    catch(IOException ioe) {
		System.out.println("getDataNodeNoChildren: IOException during read (offset="+offset_+")" + ioe.toString());		
		System.exit(0);
	    }
	}
	else {
	    if(MyFrame._debug_mode > 0) {
		System.out.println("offset "+offset_+" is beyond eof");
	    }
	}    

	return return_node;

    }
    
    /**
     * function to add a new entry to the inverted file index.
     *
     * @param word_ String object for a word 
     * @param offset_ offset in datagraph of the data node containing word_
     */
    public void addInvertedFileEntry(String word_, long offset_) {
    }

    
    /**
     * function that returns the data node object whose offset is the argument offset.
     * 
     * @param offset_ offset of a node
     * @return DataNode object whose offset is the argument offset
     */
    public DataNode getDataNode(long offset_) {
	DataNode return_node = null;
	if(_data_file == null) {
	    try {
		_data_file = new RandomAccessFile(_data_file_name, "r");
	    }
	    catch(FileNotFoundException fnfe) {
		System.out.println("FileNotFoundException during data file creation " + fnfe.toString());
		System.exit(0);
	    }
	    catch(IOException ioe) {
		System.out.println("IOException during data file creation " + ioe.toString());
		System.exit(0);
	    }
	}
	

	if(offset_ < _eof_offset) {
	    try {
		_data_file.seek(offset_);
	    }
	    catch(IOException ioe) {
		System.out.println("IOException in seek() " + ioe.toString());
		System.exit(0);
	    }
	    try {
		return_node = new DataNode(new String(""));

		int num_parents = _data_file.readInt(); // num_parents
		if(MyFrame._debug_mode > 8) {
		    System.out.println("Num parents = "+num_parents);
		}
		for(int c=0; c<num_parents; c++) {
 		    long parent_offset = _data_file.readLong(); // parent_offset
		    return_node.addParentOffset(parent_offset);
		}

		int path = _data_file.readInt();
		if(MyFrame._debug_mode > 8) {
		    System.out.println("Path = "+path);
		}
		return_node.setPath(path);

		String label = _data_file.readUTF();
		if(MyFrame._debug_mode > 8) {
		    System.out.println("Label = "+label);
		}
		return_node.setLabel(label);

		String data = _data_file.readUTF();
		if(MyFrame._debug_mode > 8) {
		    System.out.println("Data = "+data);
		}
		return_node.setData(data);

		//return_node.setChildrenLocation(_data_file.getFilePointer());

		int num_children = _data_file.readInt(); // num_children
		if(MyFrame._debug_mode > 8) {
		    System.out.println("Num children = "+num_children);
		}
		long start_time = System.currentTimeMillis();		
		for(int c=0; c<num_children; c++) {
 		    long child_offset = _data_file.readLong(); // child_offset
		    return_node.addChildOffset(child_offset);
		}
		long end_time = System.currentTimeMillis();
		//if(MyFrame._debug_mode > 8) {
		System.out.println("getDataNode: Time to read "+num_children+ " offsets = "+(end_time-start_time) +" milliseconds");
		//}
	    }
	    catch(IOException ioe) {
		System.out.println("getDataNode: IOException during read (offset="+offset_+")" + ioe.toString());		
		System.exit(0);
	    }
	}
	else {
	    if(MyFrame._debug_mode > 0) {
		System.out.println("offset "+offset_+" is beyond eof");
	    }
	}    

	return return_node;
    }

    /**
     * function that parses an XML file.
     * 
     * @param data_file_name_ name of file to be parsed
     */
    private void parseFile(String data_file_name_) {
	SAXParserFactory factory = SAXParserFactory.newInstance(); 
	factory.setValidating(false); // Thomas: This does not seem to work
	factory.setNamespaceAware(true);
	//factory.setIgnoringComments(true); // Thomas
	//factory.setIgnoreElementContentWhitespace(true);
	//factory.setCoalescing(true);
	//factory.setExpandEntityReferences(false); // Thomas

	try {
	    MyHandler handler = new MyHandler(this);
	    SAXParser parser = factory.newSAXParser();
	    parser.parse(new File(data_file_name_), handler);
	    System.out.println("Number of cross edges = " + _num_cross_edges);
	} 
	catch (SAXParseException spe) {
	    // Error generated by the parser
	    if(MyFrame._debug_mode > 0) {
		System.out.println ("\n** Parsing error" 
				    + ", line " + spe.getLineNumber ()
				    + ", uri " + spe.getSystemId ());
		System.out.println("   " + spe.getMessage() );
	    }
	    // Use the contained exception, if any
	    Exception  x = spe;
	    if (spe.getException() != null)
		x = spe.getException();
	    x.printStackTrace();
	} 
	catch (SAXException sxe) {
	    // Error generated by this application
	    // (or a parser-initialization error)
	    Exception  x = sxe;
	    if (sxe.getException() != null)
		x = sxe.getException();
	    x.printStackTrace();
	} 
	catch (ParserConfigurationException pce) {
	    // Parser with specified options can't be built
	    pce.printStackTrace();
	} 
	catch (IOException ioe) {
	    // I/O error
	    ioe.printStackTrace();
	}
    } // private function parseFile

    /** 
     * function that returns the children of node whose offset is the argument offset.
     * 
     * @param src_offset_ offset of a data node
     * @return Vector of offsets of nodes that are children of the node at offset src_offset_
     */
    public FileVector getOutEdges(long src_offset_) { // returns Vector<Long>
	DataNode src_node = this.getDataNodeNoChildren(src_offset_);
	FileVector child_offsets = src_node.getChildOffsetsOnFile();
	return child_offsets;
    }

    /**
     * function that returns the next free offset in the datagraph file.
     * 
     * @return the next free offset in the datagraph file
     */
    public long getNextFreeOffset() {
	return _eof_offset;
    }

    /**
     * function that returns the offset of the root data node.
     *
     * @return offset of root data node
     */
    public Long getRoot() {
	return new Long(_root_offset);
    }    

    /**
     * function that sets the root data node offset.
     *
     * @param offset_ offset of root data node
     */
    public void setRoot(long offset_) {
	_root_offset = offset_;
    }   

    /** 
     * function to check if a dataguide node exists for the argument path string
     * 
     * @param path_ a path string 
     * @return true if dataguide node exists for argument path string
     *         false otherwise
     */
    public boolean dataguideNodeExists(String path_) {
	return _path_dgindex_map.containsKey(path_);
    }

    /**
     * function to add a new dataguide node.
     *
     * @param dg_node_ dataguide node to be added
     */
    public void addDataguideNode(DataguideNode dg_node_) {
	if(MyFrame._debug_mode  > 8) {
	    System.out.println("Adding dataguide node for path = "+dg_node_.getPath()+", dist = "+dg_node_.getDistance());
	}
	int index = _dg_nodes.size();
	_dg_nodes.addElement(dg_node_);
	_dg_out_edges.addElement(new Vector());
	_dg_in_edges.addElement(new Vector());
	_path_dgindex_map.put(dg_node_.getPath(), new Integer(index));
    }
    /**
     * function that returns the index of dataguide node corresponding to the argument path string.
     * 
     * @param path_ a path string
     * @return index of dataguide node corresponding to the argument path string
     */
    public int getDGIndex(String path_) {
	return ((Integer)(_path_dgindex_map.get(path_))).intValue();
    }
    
    /**
     * function to set the root of the dataguide.
     *
     * @param index_ index of root of dataguide
     */
    public void setDataguideRoot(int index_) {
	_root_dg_node_index = index_;
    }

    /**
     * function that returns the root dataguide node
     *
     * @return DataguideNode object that is the root of the dataguide
     */
    public DataguideNode getDataguideRoot() {
	DataguideNode curr_dgnode = (DataguideNode)(_dg_nodes.elementAt(_root_dg_node_index));
	return curr_dgnode;
    }

    /**
     * function to add an edge between two dataguide nodes.
     *
     * @param par_index_ index of parent dataguide node
     * @param curr_index_ index of child dataguide noe
     */
    public void addDataguideEdge(int par_index_, int curr_index_) {	
	DataguideNode curr_dgnode = (DataguideNode)(_dg_nodes.elementAt(curr_index_));
	Vector out_edges = (Vector)(_dg_out_edges.elementAt(par_index_));
	if(!(out_edges.contains(curr_dgnode)))
	    out_edges.addElement(curr_dgnode);
	
	DataguideNode par_dgnode = (DataguideNode)(_dg_nodes.elementAt(par_index_));
	out_edges = (Vector)(_dg_in_edges.elementAt(curr_index_));
	if(!(out_edges.contains(par_dgnode)))
	    out_edges.addElement(par_dgnode);
    }

    /**
     * function that returns a dataguide node given an index.
     *
     * @param index_ index of a dataguide node
     * @return DataguideNode whose index is index_
     */
    public DataguideNode getDataguideNode(int index_) {
	DataguideNode dgnode = (DataguideNode)(_dg_nodes.elementAt(index_));
	return dgnode;
    }

    
    /**
     * function that returns all dataguide nodes.
     * 
     * @return Vector containing all dataguide nodes
     */
    public Vector getDataguideNodes() {
	return _dg_nodes;
    }
    
    /**
     * function that returns all dataguide out edges.
     *
     * @return Vector of Vectors, where each component Vector contains the indices of dataguide nodes
     *         that have are connected by an edge from the current node
     */
    public Vector getDataguideOutEdges() {
	return _dg_out_edges;
    }

    public Vector getDataguideInEdges() {
	return _dg_in_edges;
    }

    public int getNumDGNodes() {
	return _dg_nodes.size();
    }
}

