import java.lang.*;
import java.util.*;
import java.awt.geom.*;
import java.awt.*;

import edu.umd.cs.jazz.*;
import edu.umd.cs.jazz.util.*;
import edu.umd.cs.jazz.event.*;
import edu.umd.cs.jazz.component.*;

public class MyDataNode {
    static public final double INTER_NODE_SPACING_X = VDataguide.INTER_NODE_SPACING_X;
    static public final double INTER_NODE_SPACING_Y = VDataguide.INTER_NODE_SPACING_Y;
    static private final double ABS_PEN_WIDTH = VDataguide.ABS_PEN_WIDTH;

    /**
     * constructor.
     *
     * @param my_canvas_ canvas on which to draw the data node
     * @param node_ Jazz object 
     * @param sticky_node_ not being used
     * @param data_graph_ data graph
     * @param data_offset_ offset of the data node to be displayed
     * @param bounds_ bounding box within which to draw the data node
     * @param font_size_ size of the font that is readable at the current magnification level
     * @param camera_mag_ current magnification level
     */
    MyDataNode(ZCanvas my_canvas_, VDataguide node_, ZVisualLeaf sticky_node_, DataGraph data_graph_, 
	       Long data_offset_, 
	       Rectangle2D bounds_, int font_size_, double camera_mag_) {
	double bounding_box_top_left_x = bounds_.getX();
	double bounding_box_top_left_y = bounds_.getY();
	double bounding_box_width = bounds_.getWidth();
	double bounding_box_height = bounds_.getHeight();

	// count number of words in text
	DataNode data_node = data_graph_.getDataNodeNoChildren(data_offset_.longValue());
	String text = data_node.getData();
	int word_count = 0;
	if(text != null) {
	    StringTokenizer st = new StringTokenizer(text);
	    while(st.hasMoreTokens()) {
		String curr_word = st.nextToken();
		word_count++;
	    }
	}
		
	if(MyFrame._debug_mode > 7) {
	    System.out.println("~~~~~~~~~~~~~~~~~~~");
	    System.out.println("Text (data) string associated with node in XML graph = ");
	    System.out.println(text);
	}
	
	double text_box_width = bounding_box_width-2*INTER_NODE_SPACING_X/camera_mag_;
	double text_box_top_left_x = (bounding_box_top_left_x + INTER_NODE_SPACING_X/camera_mag_); 
	double text_box_height = bounding_box_height-2*INTER_NODE_SPACING_Y/camera_mag_;
	double text_box_top_left_y = (bounding_box_top_left_y + INTER_NODE_SPACING_Y/camera_mag_); 
	
	if(MyFrame._debug_mode > 8) {
	    System.out.println("Bounding Box Height = " + bounding_box_height);
	    System.out.println("Text Box Height = " + text_box_height);
	    System.out.println("Bounding Box Width = " + bounding_box_width);
	    System.out.println("Text Box Width = " + text_box_width);
	}
	
	if(word_count > 0) {
	    MyTextBox my_text_box = new MyTextBox(my_canvas_, node_, sticky_node_, text_box_top_left_x, text_box_top_left_y, 
						  text_box_width, text_box_height, text, camera_mag_);
	    text_box_height = my_text_box.getTextBoxHeight();
	}
	else {
	    text_box_height = 0;
	}
	
	// find how many attributes can be displayed
	String sample_label = "SPEAKERSPE"; //"WGglMyBm#/";
	int label_len = sample_label.length();       
	Font sample_label_font = new Font(null, Font.PLAIN, font_size_);
	ZLabel sample_label_obj = new ZLabel(sample_label, sample_label_font);
	sample_label_obj.setPenColor(Color.black);
	ZBounds sample_label_bounds = sample_label_obj.getBounds();
	double sample_label_width = sample_label_bounds.getWidth();///camera_mag_;
	double sample_label_height = sample_label_bounds.getHeight();///camera_mag_;
	double avg_char_width = sample_label_width/label_len;
	double avg_char_height = sample_label_height;

	int num_displayable_children = (int)((bounding_box_height-2*INTER_NODE_SPACING_Y/camera_mag_-text_box_height)/(sample_label_height+INTER_NODE_SPACING_Y));
	FileVector children = data_graph_.getOutEdges(data_offset_.longValue());
	int num_children = children.size();
	int num_child_boxes = ((num_displayable_children < num_children)? num_displayable_children:num_children);
	
	double attr_box_top_left_y;
	if(text_box_height > 0) {
	    // draw bounding box for text
	    attr_box_top_left_y = bounding_box_top_left_y+bounding_box_height-INTER_NODE_SPACING_Y/camera_mag_
		-num_child_boxes*(sample_label_height + INTER_NODE_SPACING_Y/camera_mag_);
	    ZPolyline text_bbox = new ZPolyline();
	    text_bbox.add(text_box_top_left_x, text_box_top_left_y);
	    text_bbox.add(text_box_top_left_x+text_box_width, text_box_top_left_y);
	    text_bbox.add(text_box_top_left_x+text_box_width, attr_box_top_left_y);
	    text_bbox.add(text_box_top_left_x, attr_box_top_left_y);
	    text_bbox.add(text_box_top_left_x, text_box_top_left_y);
	    text_bbox.setPenColor(Color.black);
	    text_bbox.setPenWidth(ABS_PEN_WIDTH/camera_mag_);
	    node_.addVisualComponent(text_bbox);
	    attr_box_top_left_y = attr_box_top_left_y + INTER_NODE_SPACING_Y/camera_mag_;
	}
	else {	    
	    attr_box_top_left_y = bounding_box_top_left_y + INTER_NODE_SPACING_Y/camera_mag_;
	    sample_label_height = (bounding_box_height-(num_child_boxes+1)*INTER_NODE_SPACING_Y/camera_mag_)/num_child_boxes;
	}

	// Display as many attributes as space permits
	int child_num = 0;
 	while(child_num < num_child_boxes) {
	    Long child_offset = (Long)(children.elementAt(child_num));
	    DataNode child = (DataNode)(data_graph_.getDataNodeNoChildren(child_offset.longValue()));
	    String child_label_string = child.getLabel();
	    
	    MyLabelBox attr_visual_obj = new MyLabelBox(my_canvas_, node_, sticky_node_, text_box_top_left_x, attr_box_top_left_y,
							text_box_width, sample_label_height, 
							child_label_string, avg_char_width, avg_char_height, 
							font_size_, camera_mag_);
	    attr_box_top_left_y = attr_box_top_left_y + (sample_label_height+INTER_NODE_SPACING_Y/camera_mag_);
	    child_num++;
	}	
    }
}
