import java.io.*;
import java.util.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

import edu.umd.cs.jazz.*;
import edu.umd.cs.jazz.event.*;
import edu.umd.cs.jazz.util.*;
import edu.umd.cs.jazz.component.*;
import javax.swing.*;

public class MyFrame extends JFrame {
    private ZSceneGraphTreeView treeView = null;
    protected ZCanvas _my_canvas; 
    static public int _debug_mode = 0;
    static public boolean _scale_text = false; 
    static public boolean _smooth_wrap = false;

    public MyFrame() {
	// gracefully exit on closing of window
	addWindowListener(new WindowAdapter() {
	    public void windowClosing(WindowEvent e) {
		System.exit(0);
	    }
	}); 

	Container cp = getContentPane();
	setBounds(100, 100, 400, 350); //1000, 800);
	JMenuBar my_menu_bar = createMyMenuBar();
	setJMenuBar(my_menu_bar);
	_my_canvas = new ZCanvas();
	cp.add(_my_canvas, BorderLayout.CENTER);
	validate();
	setVisible(true);
    }
    
    private JMenuBar createMyMenuBar() {
	JMenuBar my_menu_bar = new JMenuBar();
	
	// The 'File' Menu
	JMenu file_menu = new JMenu("File");
	file_menu.setMnemonic('F');
	my_menu_bar.add(file_menu);
	
	// The 'Open XML' Menu Item
	JMenuItem open_file_menu_item = new JMenuItem("Open XML File");
	open_file_menu_item.setMnemonic('X');
	open_file_menu_item.addActionListener(new ActionListener() {
	    public void actionPerformed(ActionEvent ae) {
		openXMLFile();
	    }
	});
	file_menu.add(open_file_menu_item);
	

	// The 'Index XML' Menu Item
	JMenuItem index_file_menu_item = new JMenuItem("Index XML File");
	index_file_menu_item.setMnemonic('I');
	index_file_menu_item.addActionListener(new ActionListener() {
	    public void actionPerformed(ActionEvent ae) {
		indexXMLFile();
	    }
	});
	file_menu.add(index_file_menu_item);
	

	// The 'Index XML' Menu Item
	JMenuItem clear_screen_menu_item = new JMenuItem("Clear Screen");
	clear_screen_menu_item.setMnemonic('C');
	clear_screen_menu_item.addActionListener(new ActionListener() {
	    public void actionPerformed(ActionEvent ae) {
		clearScreen();
	    }
	});
	file_menu.add(clear_screen_menu_item);
	
	// The 'Quit' Menu Item
	JMenuItem quit_menu_item = new JMenuItem("Quit");
	quit_menu_item.setMnemonic('Q');
	quit_menu_item.addActionListener(new ActionListener() {
	    public void actionPerformed(ActionEvent ae) {
		if(MyFrame._debug_mode > 8) {
		    System.out.println("garbage collector running");
		}
		System.gc();
		if(MyFrame._debug_mode > 8) {
		    System.out.println("garbage collection finished");	
		}
		System.exit(0);
	    }
	});
	file_menu.add(quit_menu_item);

	// The 'View' Menu
	JMenu view_menu = new JMenu("View");
	view_menu.setMnemonic('V');
	my_menu_bar.add(view_menu);
	
	// The 'Debug View' Menu Item
	JMenuItem debug_menu_item = new JMenuItem("Debug View");
	debug_menu_item.setMnemonic('D');
	debug_menu_item.addActionListener(new ActionListener() {
	    public void actionPerformed(ActionEvent ae) {
		showTreeView();
	    }
	});
	view_menu.add(debug_menu_item);

	return my_menu_bar;
    }
    
    private void openXMLFile() {
	File file_to_open = null;
	ExtensionFileFilter file_name_filter = new ExtensionFileFilter(
								       new String("xml"), 
								       "XML & DTD files");
	JFileChooser file_chooser = new JFileChooser("");
	file_chooser.addChoosableFileFilter(file_name_filter);
	int retval = file_chooser.showDialog(_my_canvas, "Open XML File");
	if (retval == JFileChooser.APPROVE_OPTION) {
	    file_to_open = file_chooser.getSelectedFile();
	}
	if(file_to_open == null) return;
	
	this.clearScreen();

	String file_name_to_open = file_to_open.getAbsolutePath();
	if(MyFrame._debug_mode > 0) {
	    System.out.println("Opening XML file "+file_name_to_open);
	}
	StringTokenizer tokenizer = new StringTokenizer(file_name_to_open, new String("\\/"));
	String file_name = null;
	while(tokenizer.hasMoreTokens()) {
	    file_name = tokenizer.nextToken();
	}
	String index_directory = "."+file_name+".dir/";
	File index_directory_obj = new File(index_directory);
	if(!index_directory_obj.exists()) index_directory_obj.mkdir();

	if(_debug_mode > 0) {
	    System.out.println("Starting to create data graph on disk");
	}
	long start_time = System.currentTimeMillis();
	DataGraph data_graph = new DataGraph(index_directory, file_name, true);
	long end_time = System.currentTimeMillis();
	if(_debug_mode > 0) {
	    System.out.println("Created data graph; Time taken = "+(end_time-start_time)+ " milliseconds");
	}
	if(_debug_mode > 5) {
	    System.out.println("-------------------------");
	}
	if(_debug_mode > 0) {
	    System.out.println("Starting to create dataguide");
	}
	start_time = System.currentTimeMillis();
	Dataguide data_guide = new Dataguide(data_graph);
	end_time = System.currentTimeMillis();
	data_guide.flush();

	if(_debug_mode > 0) {
	    System.out.println("Created dataguide; Time taken = "+(end_time-start_time)+ " milliseconds");
	}
	if(_debug_mode > 5) {
	    System.out.println("-------------------------");
	}

	if(_debug_mode > 0) {
	    System.out.println("Creating Jazz object.");
	}
	start_time = System.currentTimeMillis();
	VDataguide visual_data_guide = new VDataguide(_my_canvas, data_guide);	
	end_time = System.currentTimeMillis();
	if(_debug_mode > 0) {
	    System.out.println("Created Jazz object; Time taken = "+(end_time-start_time)+ " milliseconds");
	}
	
    }

    private void indexXMLFile() {
	File file_to_open = null;
	ExtensionFileFilter file_name_filter = new ExtensionFileFilter(
								       new String("xml"), 
								       "XML & DTD files");
	JFileChooser file_chooser = new JFileChooser("");
	file_chooser.addChoosableFileFilter(file_name_filter);
	int retval = file_chooser.showDialog(_my_canvas, "Index XML File");
	if (retval == JFileChooser.APPROVE_OPTION) {
	    file_to_open = file_chooser.getSelectedFile();
	}
	if(file_to_open == null) return;
	
	this.clearScreen();

	String file_name_to_open = file_to_open.getAbsolutePath();
	if(MyFrame._debug_mode > 0) {
	    System.out.println("Indexing XML file "+file_name_to_open);
	}
	StringTokenizer tokenizer = new StringTokenizer(file_name_to_open, new String("\\/"));
	String file_name = null;
	while(tokenizer.hasMoreTokens()) {
	    file_name = tokenizer.nextToken();
	}
	String index_directory = "."+file_name+".dir/";
	File index_directory_obj = new File(index_directory);
	if(!index_directory_obj.exists()) index_directory_obj.mkdir();

	if(_debug_mode > 0) {
	    System.out.println("Starting to create data graph on disk");
	}
	long start_time = System.currentTimeMillis();
	DataGraph data_graph = new DataGraph(index_directory, file_name, true);
	long end_time = System.currentTimeMillis();
	if(_debug_mode > 0) {
	    System.out.println("Created data graph; Time taken = "+(end_time-start_time)+ " milliseconds");
	}

	if(_debug_mode > 5) {
	    System.out.println("-------------------------");
	}
	if(_debug_mode > 0) {
	    System.out.println("Starting to create dataguide");
	}
	start_time = System.currentTimeMillis();
	Dataguide data_guide = new Dataguide(data_graph);
	end_time = System.currentTimeMillis();
	data_guide.flush();

	if(_debug_mode > 0) {
	    System.out.println("Created dataguide; Time taken = "+(end_time-start_time)+ " milliseconds");
	}
	if(_debug_mode > 5) {
	    System.out.println("-------------------------");
	}
    }

    public void showTreeView() {
        if (treeView == null) {
            treeView = new ZSceneGraphTreeView(_my_canvas);
            treeView.addWindowListener(new WindowAdapter() {
                public void windowDeactivated(WindowEvent e){
                    treeView = null;
                }
                public void windowClosed(WindowEvent e) {
                    treeView = null;
                }
            });
            treeView.pack();
            treeView.show();
        }
    }    
    
    private void clearScreen() {
	if(_debug_mode > 0) {
	    System.out.println("Explicitly calling garbage collector...");
	}
	System.gc();
	if(_debug_mode > 0) {
	    System.out.println("Garbage collection done.");
	}
	Container cp = getContentPane();
	cp.remove(_my_canvas);
	_my_canvas = new ZCanvas();
	cp.add(_my_canvas, BorderLayout.CENTER);
	validate();
	setVisible(true);

    }

    public ZCanvas getCanvas() {
	return _my_canvas;
    }

    public static void main(String args[]) {
	int num_args = args.length; // -- Thomas, How do you find the number of args in am array
	for(int argnum = 0; argnum<num_args; argnum++) {
	    String curr_arg = (String)args[argnum];
	    if(curr_arg.compareToIgnoreCase("-debug") == 0) {
		// find debug level
		argnum++;
		if(argnum < num_args) {
		    curr_arg = (String)args[argnum];
		    MyFrame._debug_mode = Integer.valueOf(curr_arg).intValue();
		}
	    }
	    else if(curr_arg.compareToIgnoreCase("-scaletext") == 0) {
		MyFrame._scale_text = true;
	    }
	    else if(curr_arg.compareToIgnoreCase("-noscaletext") == 0) {
		MyFrame._scale_text = false;
	    }
	    else if(curr_arg.compareToIgnoreCase("-smoothwrap") == 0) {
		MyFrame._smooth_wrap = true;
	    }
	    else if(curr_arg.compareToIgnoreCase("-nosmoothwrap") == 0) {
		MyFrame._smooth_wrap = false;
	    }
	}
	MyFrame my_frame = new MyFrame();
    }
}
