import java.lang.*;
import java.util.*;
import org.w3c.dom.*;
import java.io.*;

import org.xml.sax.*;

public class MyHandler extends HandlerBase {
    private Stack _node_stack = null;
    //Random _random_num_gen = null;
    private DataGraph _data_graph = null;
    private long _counter = 0;
    private int _max_depth = -1;
    private int _curr_depth = -1;
    private long _start_time = 0;
    private Random _dg_node_id_gen = new Random();

    /**
     * constructor
     *
     * @param data_graph_ Datagraph object representing the data graph being created
     */
    MyHandler(DataGraph data_graph_) {
	_node_stack = new Stack();
	//_random_num_gen = new Random();
	_data_graph = data_graph_;
    }
    
    /**
     * callback function that is called when a start tag is encountered by the parser.
     *
     * @param element_ name of the element
     * @param attributes_ attributes of the element
     */
    public void startElement(String element_, AttributeList attributes_) {
	if(_counter == 0) {
	    _start_time = System.currentTimeMillis();
	}
	    
	if(_counter%30000 == 0) {
	    long time_taken = System.currentTimeMillis() - _start_time;
	    System.out.println("Number of Elements Parsed = "+_counter+"; Elapsed Time = "+time_taken+" milliseconds");
	}
	_counter++;
	
	// add data node
	_curr_depth++;
	if(_curr_depth > _max_depth) {
	    if(_max_depth >= 0) _data_graph.createParentOutStream(new Integer(_max_depth));
	    _max_depth = _curr_depth;
	}

	DataNode curr_data_node = new DataNode(/*id_string, idref_string, */element_);
	_node_stack.push(curr_data_node);

	/*
	for(int i = 0; i < attributes_.getLength(); i++) {
	    String name = attributes_.getName(i);
	    if(name.equalsIgnoreCase("id")) continue;
	    else if(name.equalsIgnoreCase("idref")) continue;
	    else {
		long attr_node_offset = _data_graph.getNextFreeOffset();
		String type = attributes_.getType(i);
		String value = attributes_.getValue(i);	  
		String attr_id_string = new String("Data." + _random_num_gen.nextInt());
		DataNode my_data_node = new DataNode(attr_id_string, "", name);		
		my_data_node.setData(value);
		curr_data_node.addChildOffset(attr_node_offset);
		my_data_node.addParentOffset(-1);

		if(MyFrame._debug_mode > 5) {
		    System.out.println("Adding attr node "+my_data_node.getId()+
				       ", label="+my_data_node.getLabel()+
				       ", idref="+my_data_node.getIdref());
		}
		_data_graph.addDataNode(attr_node_offset, my_data_node);
	    }
	}	
	*/
    }

    /**
     * callback function that is called when an end tag is encountered by the parser.
     *
     * @param element_ name of the element
     */
    public void endElement(String element_) {
	// compute the current path
	Enumeration e = _node_stack.elements();
	StringBuffer curr_path_buf = new StringBuffer("");
	String curr_label = new String("");
	int curr_dist = 0;
	while(e.hasMoreElements()) {
	    DataNode curr_stk_node = (DataNode)(e.nextElement());
	    if(e.hasMoreElements()) {
		curr_path_buf.append(curr_stk_node.getLabel()+".");
		curr_dist++;
	    }
	    else {
		curr_label = curr_stk_node.getLabel();
		curr_path_buf.append(curr_label);
	    }
	}
	
	// remember the offset at which the current node is to be written	
	long curr_file_offset = _data_graph.getNextFreeOffset();
	DataNode curr_node = (DataNode)(_node_stack.pop());

	// compute the parent path
	e = _node_stack.elements();
	StringBuffer par_path_buf = new StringBuffer("");
	String par_label = new String("");
	int par_dist = 0;
	while(e.hasMoreElements()) {
	    DataNode curr_stk_node = (DataNode)(e.nextElement());
	    if(e.hasMoreElements()) {
		par_path_buf.append(curr_stk_node.getLabel()+".");
		par_dist++;
	    }
	    else {
		par_label = curr_stk_node.getLabel();
		par_path_buf.append(par_label);
	    }
	}
	
	// parent and current node paths
	String curr_path = curr_path_buf.toString();
	String par_path = par_path_buf.toString();

	// add dataguide nodes and edges
	if(_data_graph.dataguideNodeExists(curr_path) == false) {
	    int dgindex = _data_graph.getNumDGNodes();
	    if(MyFrame._debug_mode > 5) {
		System.out.println("curr path = "+curr_path);
		System.out.println("parent path = "+par_path);
	    }
	    String dg_node_id = new String("Schema." + _dg_node_id_gen.nextInt());
	    DataguideNode curr_dgnode = new DataguideNode(dgindex, _data_graph.getIndexDirectory(), dg_node_id, curr_label);
	    curr_dgnode.setPath(curr_path, curr_dist);
	    _data_graph.addDataguideNode(curr_dgnode);	    
	}
	int curr_dgindex = _data_graph.getDGIndex(curr_path);
	DataguideNode curr_dgnode = _data_graph.getDataguideNode(curr_dgindex);
	curr_node.setPath(curr_dgindex);
	curr_dgnode.addDataNode(new Long(curr_file_offset));

	if(par_path.equals(new String(""))) {
	    if(MyFrame._debug_mode > 5) {
		System.out.println("Setting dg root: label = "+curr_path);
	    }
	    _data_graph.setDataguideRoot(curr_dgindex);
	}
	else {
	    if(_data_graph.dataguideNodeExists(par_path) == false) {
		int dgindex = _data_graph.getNumDGNodes();
		String dg_node_id = new String("Schema." + _dg_node_id_gen.nextInt());
		DataguideNode par_dgnode = new DataguideNode(dgindex, _data_graph.getIndexDirectory(), dg_node_id, par_label);
		par_dgnode.setPath(par_path, par_dist);
		_data_graph.addDataguideNode(par_dgnode);
	    }
	    int par_dgindex = _data_graph.getDGIndex(par_path);
	    _data_graph.addDataguideEdge(par_dgindex, curr_dgindex);
	}

	
	if(_node_stack.empty() == false) {
	    DataNode parent_node = (DataNode)(_node_stack.peek());
	    parent_node.addChildOffset(curr_file_offset);
	    curr_node.addParentOffset(-1);
	}
	else {
	    _data_graph.setRoot(curr_file_offset);
	}

	_data_graph.addDataNode(curr_node, _curr_depth);
	_curr_depth--;
    }
    
    /**
     * callback function that is called when the parser encounters data within an element.
     *
     * @param ch_ array of characters 
     * @param start_ start position in array where the element's data begins
     * @param length_ number of characters in the element's data
     */
    public void characters(char[] ch_, int start_, int length_) {	
	StringBuffer char_buffer = new StringBuffer();
	for(int i=start_; i<length_; i++) {
	    if(ch_[i] == '\n') char_buffer.append(' ');
	    else char_buffer.append(ch_[i]);
	}
	String curr_data = char_buffer.toString();	
	DataNode curr_data_node = (DataNode)(_node_stack.peek());
	curr_data_node.setData(curr_data.trim());
    }

    /**
     * callback function that is called by the parser when the end of the document is reached.
     */

    public void endDocument() {
	long time_taken = System.currentTimeMillis() - _start_time;
	System.out.println("Processed Element Count = "+_counter+", Elapsed Time = "+time_taken+" milliseconds");
    }

}
