import java.lang.*;
import java.util.*;
import java.awt.geom.*;
import java.awt.*;

import edu.umd.cs.jazz.*;
import edu.umd.cs.jazz.util.*;
import edu.umd.cs.jazz.event.*;
import edu.umd.cs.jazz.component.*;

public class TimerThread extends Thread {
    private boolean _time_to_act = false;
    private LinkedList _to_do_list = null;
    private int _counter = 0;
    private VDataguide _node = null;
    
    static public final long CYCLE_TIME = 200;
    static public final int MAX_PROCASTINATE_COUNT = 10;

    class ZoomEvent {
	private long _curr_time = 0;
	
	ZoomEvent(long curr_time_) {
	    _curr_time = curr_time_;
	}
    }

    /**
     * constructor.
     *
     * @param node_ Jazz object 
     */
    TimerThread(VDataguide node_) {
	_node = node_;
	_to_do_list = new LinkedList();
	_counter = 0;
    }

    /**
     * 
     */
    public void run() {
	while(true) {
	    int start_size = _to_do_list.size();
	    // sleep for some time
	    try {
		sleep(CYCLE_TIME);
	    }
	    catch(InterruptedException ie) {

	    }
	    int end_size = _to_do_list.size();

	    // if no zoom event occurred while I slept, perform events in the todo list
	    if(start_size == end_size) {
		if(start_size > 0) {
		    Object next_event = _to_do_list.removeFirst();
		    _time_to_act = true;		    
		    if(MyFrame._debug_mode > 5) {
			System.out.println("*****TimerThread: Low input rate*****");
		    }
		    _node.timer_expired(next_event);
		}
		_counter = 0;	
	    }
	    // else if counter < MAX_PROCASTINATE_COUNT, append new events to the todo list and increment counter
	    else if(_counter < MAX_PROCASTINATE_COUNT) {
		_counter++;
	    }
	    // else perform events in the todo list
	    else {
		Object next_event = _to_do_list.removeFirst();
		_time_to_act = true;
		if(MyFrame._debug_mode > 5) {
			System.out.println("*****TimerThread: High input rate; Forcing update*****");
		}
		_node.timer_expired(next_event);
		_counter = 0;		
	    }	    
	}
    }

    /**
     * function to queue a zoom event when the input rate is high.
     *
     * @param curr_time_ time when the zoom event was generated
     */
    public synchronized void queueEvent(Object curr_event_) {
	_to_do_list.add(curr_event_);
    }

    /**
     * function that returns true if queued zoom events have to be processed.
     *
     * @return true if it's time to process queued zoom events
     *         false otherwise
     */
    public synchronized boolean isTime() {
	return _time_to_act;
    }

    /**
     * function to reset the "time to process events" flag.
     */
    public synchronized void resetTimeFlag() {
	_time_to_act = false;
    }
}
